/*****************************************************************************/
/*                                                                           */
/*    TIMWASTE.C -- CPU performance test program                             */
/*                                                                           */
/*    Written by Walter Oney                                                 */
/*                                                                           */
/*****************************************************************************/

#define WIN32_LEAN_AND_MEAN      // excludes mmsystem.h, among others
#include <windows.h>
#include <stdlib.h>
#include <string.h>

/* We use timeGetTime on Win16 to get better (and repeatable) resolution of
   timing intervals. On Win32, use the high-resolution timer provided by
   QueryPerformanceCounter. */

#ifndef WIN32
#include <mmsystem.h>            // for timeGetTime
#else
static DWORD timeGetTime(void);
#endif

#include "timwaste.h"

static HINSTANCE hInst;

static int howsort;
static int numelem = 1000;

LRESULT EXPORT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
BOOL EXPORT CALLBACK OptionsDlg(HWND, UINT, WPARAM, LPARAM);

static void sort(void);
static int compare(const LPINT left, const LPINT right);
static void interchange(LPINT left, LPINT right);

typedef int (*SORTPROC)(const void *, const void *);

/*****************************************************************************/

int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrev, LPSTR lpCmd,
   int nShow)
   {                             // WinMain
   HWND hwnd;
   MSG msg;
   WNDCLASS wc;

   hInst = hInstance;
   memset(&wc, 0, sizeof(wc));
   wc.lpszClassName = "timwaste";
   wc.hInstance = hInstance;
   wc.lpfnWndProc = WndProc;
   wc.hCursor = LoadCursor(NULL, IDC_ARROW);
   wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
   wc.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1);
   wc.lpszMenuName = MAKEINTRESOURCE(IDR_MENU1);

   if (!RegisterClass(&wc))
      return 0;
   hwnd = CreateWindow("timwaste", "CPU Performance Test Program",
      WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
      0, NULL, hInstance, NULL);
   if (!hwnd)
      return 0;

   ShowWindow(hwnd, nShow);

   while (GetMessage(&msg, 0, 0, 0))
      {                          // process messages
      TranslateMessage(&msg);
      DispatchMessage(&msg);
      }                          // process messages
   return msg.wParam;
   }                             // WinMain

/*****************************************************************************/

LRESULT EXPORT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam,
   LPARAM lParam)
   {                             // WndProc
   switch (msg)
      {                          // process message

   /*------------------------------------------------------------------------*/

   case WM_COMMAND:
      switch (LOWORD(wParam))
         {                       // process command

      case ID_FILE_EXIT:
         SendMessage(hwnd, WM_CLOSE, 0, 0);
         return 0;

      case ID_FILE_SORT:
         sort();
         return 0;

      case ID_OPTIONS:
         {                       // ID_OPTIONS
         DLGPROC dlgproc = (DLGPROC) MakeProcInstance((FARPROC) OptionsDlg,
            hInst);
         DialogBox(hInst, MAKEINTRESOURCE(IDD_Options), hwnd, dlgproc);
         FreeProcInstance((FARPROC) dlgproc);
         return 0;
         }                       // ID_OPTIONS
         }                       // process command
      return 0;

   /*------------------------------------------------------------------------*/

   case WM_DESTROY:
      PostQuitMessage(0);
      return 0;

      }                          // process message
   return DefWindowProc(hwnd, msg, wParam, lParam);
   }                             // WndProc

/*****************************************************************************/

BOOL EXPORT CALLBACK OptionsDlg(HWND hdlg, UINT msg, WPARAM wParam,
   LPARAM lParam)
   {                             // OptionsDlg
   switch (msg)
      {                          // process message

   /*------------------------------------------------------------------------*/

   case WM_INITDIALOG:
      CheckRadioButton(hdlg, IDC_LinearExchange, IDC_Qsort,
         IDC_LinearExchange + howsort);
      SetDlgItemInt(hdlg, IDC_NumElements, numelem, FALSE);
      return TRUE;

   /*------------------------------------------------------------------------*/

   case WM_COMMAND:
      {                          // WM_COMMAND
      UINT wID = LOWORD(wParam);
      switch (wID)
         {                       // process control notification

      case IDOK:
         {                       // IDOK
         BOOL okay;
         int newnum = GetDlgItemInt(hdlg, IDC_NumElements, &okay, FALSE);
         if (!okay)
            {                    // invalid element count
            MessageBeep(MB_ICONHAND);
            SetFocus(GetDlgItem(hdlg, IDC_NumElements));
            SendDlgItemMessage(hdlg, IDC_NumElements, EM_SETSEL, 0,
               MAKELPARAM(0, -1));
            break;
            }                    // invalid element count
         for (howsort = 0; howsort < 4; ++howsort)
            if (IsDlgButtonChecked(hdlg, IDC_LinearExchange + howsort))
               break;
         numelem = newnum;
         EndDialog(hdlg, 1);
         break;
         }                       // IDOK

      case IDCANCEL:
         {                       // IDCANCEL
         EndDialog(hdlg, 0);
         break;
         }                       // IDCANCEL

         }                       // process control notification
      return TRUE;
      }                          // WM_COMMAND

      }                          // process message
   return FALSE;
   }                             // OptionsDlg

/*****************************************************************************/

/* Element compare routine used for all flavors of sort: */

static int compare(const LPINT left, const LPINT right)
   {                             // compare
   return *left - *right;
   }                             // compare

/*****************************************************************************/

/* Element interchange routine used for linear exchange and Shell sorts: */

static void interchange(LPINT left, LPINT right)
   {                             // switch
   int temp = *left;
   *left = *right;
   *right = temp;
   }                             // switch

/*****************************************************************************/

/* Timer routine for Win32 systems: */

#ifdef WIN32

static DWORD timeGetTime(void)
   {                             // timeGetTime
   static DWORD freq;            // timer frequency
   LARGE_INTEGER curtime;

   if (!freq)
      {                          // determine timer frequency
      QueryPerformanceFrequency(&curtime);
      if (curtime.HighPart)
         {                       // timer is too fast
         MessageBox(NULL, "High resolution timer counts too quickly for"
            " single-width arithmetic. Use LARGEINT SDK sample.",
            "Warning", MB_OK | MB_ICONEXCLAMATION);
         freq = 1;
         }                       // timer is too fast
      else
         freq = curtime.LowPart / 1000; // i.e., ticks per millisecond
      }                          // determine timer frequency
   QueryPerformanceCounter(&curtime);
   return curtime.LowPart / freq;
   }                             // timeGetTime

#endif // WIN32

/*****************************************************************************/

/* SORT is the routine that demonstrates the various timing tools. You
   can parameterize a sort with the Options dialog, which allows you to
   specify the number of elements to be sorted and the method to employ. */

static void sort(void)
   {                             // sort
   int i;                        // loop control
   LPINT array;                  // array to sort
   HGLOBAL hmem;                 // array memory handle
   DWORD start, end;             // start & end times
   char msg[128];                // report string
   HCURSOR hold;                 // original cursor
   static char *sortname[] = {"Linear exchange sort",
      "Linear exchange sort with inline compare/interchange",
      "Shell sort", "Library qsort"};

/* Allocate and initialize an array of the specified size and populate it
   with pseudo-random integers. */

   if (!(hmem = GlobalAlloc(GMEM_MOVEABLE, numelem * sizeof(int))))
      return;
   if (!(array = GlobalLock(hmem)))
      {                          // can't lock memory block
      GlobalFree(hmem);
      return;
      }                          // can't lock memory block

   for (i = 0; i < numelem; ++i)
      array[i] = rand();

/* Sort the table using the specified method. */

   hold = SetCursor(LoadCursor(NULL, IDC_WAIT));
   start = timeGetTime();
   switch (howsort)
      {                          // perform sort

   /*------------------------------------------------------------------------*/

   /* A linear exchange sort, sometimes also called a bubble sort, is the
      first one most of us learn because it's so easy to explain. It performs
      in o(n**2), which makes it unsuitable for all but the smallest tables. */

   case 0:
      {                          // linear exchange sort
      int j;
      for (i = 1; i < numelem; ++i)
         for (j = i-1; j >= 0; --j)
            if (compare(array+j, array+j+1) > 0)
               interchange(array+j, array+j+1);
            else
               break;
      break;
      }                          // linear exchange sort

   /*------------------------------------------------------------------------*/

   /* One misguided solution to the poor performance of the exchange sort
      is to put the compare and interchange routines inline. For Intel
      implementations, it's easy to do this with inline assembler, too.
      The assembler version ran about twice as fast as the C version in
      one of my benchmarks, which mostly indicates how much better hand
      coding is than the "Od" code generated by C8. */

   #ifdef _X86_

   case 1:
      {                          // inline compare/interchange
      int j;
      for (i = 1; i < numelem; ++i)
         {                       // for each element
         for (j = i-1; j >= 0; --j)
            _asm
               {                 // compare & interchange
               mov   eax, array
               mov   ebx, j
               mov   ecx, [eax+4*ebx]
               cmp   ecx, [eax+4+4*ebx]
               jle   next_i
               xchg  ecx, [eax+4+4*ebx]
               mov   [eax+4*ebx], ecx
               }                 // compare & interchange
      next_i:;
         }                       // for each element
      break;
      }                          // inline compare/interchange

   #else  // not _X86_

   case 1:
      {                          // inline compare/interchange
      int j;
      for (i = 1; i < numelem; ++i)
         {                       // for each element
         for (j = i-1; j >= 0; --j)
            if (array[j] > array[j+1])
               {                 // interchange elements
               int temp = array[j];
               array[j] = array[j+1];
               array[j+1] = temp;
               }                 // interchange elements
            else
               break;
         }                       // for each element
      }                          // inline compare/interchange

   #endif // _X86_

   /*------------------------------------------------------------------------*/

   /* This is a simple implementation of Shell's o(n*log(n)) sort. According
      to Knuth's Sorting and Searching, there are better choices for the "gap"
      variable, but this implementation is very easy to code and gives
      acceptable performance. */

   case 2:
      {                          // shell sort
      int j, gap;
      for (gap = numelem >> 1; gap > 0; gap >>= 1)
         for (i = gap; i < numelem; ++i)
            for (j = i-gap; j >= 0; j -= gap)
               if (compare(array+j, array+j+gap) > 0)
                  interchange(array+j, array+j+gap);
               else
                  break;
      break;
      }                          // shell sort

   /*------------------------------------------------------------------------*/

   /* The "quicksort" algorithm attributed to C.A.R. Hoare has very good
      performance and is so popular that it's been included in most runtime
      libraries. You could recode this by hand for the absolute ultimate in
      performance. */

   case 3:
      {                          // qsort
      qsort(array, numelem, sizeof(int), (SORTPROC) compare);
      break;
      }                          // qsort

      }                          // perform sort
   end = timeGetTime();

/* Report the results of the sort. Note that most performance monitors will
   tell you that the following MessageBox call is where most of the time is
   spent because they measure wall-clock time. */

   SetCursor(hold);
   wsprintf(msg, "%s of %d items took %ld ms", (LPSTR) sortname[howsort],
      numelem, end-start);
   MessageBox(NULL, msg, "Sort results", MB_OK | MB_ICONINFORMATION);

   GlobalUnlock(hmem);
   GlobalFree(hmem);
   }                             // sort
